/**
  *  \file client/screens.cc
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens.h"
#include "cpluslib/assert.h"
#include "game/ship.h"
#include "game/planet.h"
#include "game/global.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/fleet.h"
#include "int/if/userif.h"
#include "int/value.h"
#include "client/chartusr.h"
#include "game/ion.h"

/** Create blank screen Id.
    \post !isValid */
WScreenId::WScreenId()
    : screen_id(NullScreen), x(0), y(0)
{ }

/** Create screen Id for an object.
    \param screen_id Screen identifier
    \param id        Selected object on that screen. Must be valid (not checked!)
                     or zero. */
WScreenId::WScreenId(ScreenId screen_id, int id)
    : screen_id(screen_id), x(id), y(0)
{
    // FIXME: add this when all code is converted
    // ASSERT(screen_id != Starchart);
}

/** Create screen Id for map location.
    \param pt Map location */
WScreenId::WScreenId(GPoint pt)
    : screen_id(Starchart), x(pt.x), y(pt.y)
{ }

/** Get object Id. */
int
WScreenId::getId() const
{
    ASSERT(screen_id != Starchart);
    ASSERT(screen_id != NullScreen);
    return x;
}

/** Get starchart position.
    \pre getScreenId() == Starchart */
GPoint
WScreenId::getPos() const
{
    ASSERT(screen_id == Starchart);
    return GPoint(x, y);
}

bool
WScreenId::isValid() const
{
    switch (screen_id) {
     case RaceScreen:
     case Starchart:
        return true;

     case ShipScreen:
     case ShipTaskScreen:
        return current_ship_type.isValidIndex(x);

     case PlanetScreen:
     case PlanetTaskScreen:
        return current_planet_type.isValidIndex(x);

     case HistoryScreen:
        return current_histship_type.isValidIndex(x);

     case BaseScreen:
     case BaseTaskScreen:
        return current_base_type.isValidIndex(x);

     case FleetScreen:
        return current_fleet_type.isValidIndex(x);

     default:
        return false;
    }
}

string_t
WScreenId::toString()
{
    switch (screen_id) {
     case RaceScreen:
        return _("Race Screen");
     case Starchart:
        return format(_("Starchart (%d,%d)"), x, y);
     case ShipScreen:
        return getDisplayedTurn().getCurrentUniverse().getShip(x).getName(GObject::LongName);
     case PlanetScreen:
        return getDisplayedTurn().getCurrentUniverse().getPlanet(x).getName(GObject::LongName);
     case BaseScreen:
        return format(_("Starbase #%d: %s"), x, getDisplayedTurn().getCurrentUniverse().getPlanet(x).getName(GObject::PlainName));
     case HistoryScreen:
        return format(_("Ship History %d: %s"), x, getDisplayedTurn().getCurrentUniverse().getShip(x).getName(GObject::PlainName));
     case FleetScreen:
        return getFleetTitle(getDisplayedTurn().getCurrentUniverse(), x);
     case ShipTaskScreen:
        return format(_("Ship Task #%d: %s"), x, getDisplayedTurn().getCurrentUniverse().getShip(x).getName(GObject::PlainName));
     case PlanetTaskScreen:
        return format(_("Planet Task #%d: %s"), x, getDisplayedTurn().getCurrentUniverse().getPlanet(x).getName(GObject::PlainName));
     case BaseTaskScreen:
        return format(_("Starbase Task #%d: %s"), x, getDisplayedTurn().getCurrentUniverse().getPlanet(x).getName(GObject::PlainName));
     default:
        return format("<%d,%d,%d>", int(screen_id), x, y);
    }
}


static bool
trySetIndex(int index, GObjectType& type, GObjectSelection& sel)
{
    if (index == 0) {
        // No change requested; verify that current index is good.
        // (This will return failure if the type is empty.)
        return type.isValidIndex(sel.getCurrentIndex());
    } else {
        // Change requested.
        if (type.isValidIndex(index)) {
            sel.setCurrentIndex(index);
            return true;
        } else {
            return false;
        }
    }
}

/** Given an object, find its index.
    \param type Object type
    \param obj Object to find
    \return index if found, otherwise 0 */
static int
findIndex(GObjectType& type, GObject* obj)
{
    for (int i = type.findNextIndex(0); i != 0; i = type.findNextIndex(i)) {
        if (type.isValidIndex(i) && &type.getObjectByIndex(i) == obj) {
            return i;
        }
    }
    return 0;
}


/** Post a "Go To Screen" command.
    \param screen Screen number
    \param id Object Id. Can be 0.
    \return true iff request could be processed */
bool
postGoToScreen(IntUserInterfaceBinding& ui, int screen, int id)
{
    // Validate args
    bool ok;
    switch (screen) {
     case 0:
     case 4:
        // Those don't take args
        ok = (id == 0);
        break;

     case 1:
     case 11:
        // Ship Id
        ok = trySetIndex(id, current_ship_type, current_ship_selection);
        break;

     case 2:
     case 12:
        // Planet Id
        ok = trySetIndex(id, current_planet_type, current_planet_selection);
        break;

     case 3:
     case 13:
        // Base
        ok = trySetIndex(id, current_base_type, current_base_selection);
        break;

     case 6:
        // History ship
        ok = trySetIndex(id, current_histship_type, current_histship_selection);
        break;

     case 10:
        // Fleet
        ok = trySetIndex(id, current_fleet_type, current_fleet_selection);
        break;

     default:
        ok = false;
        break;
    }

    // On success, post the event
    if (ok) {
        std::auto_ptr<IntUserInterfaceEvent> event(new IntUserInterfaceEvent(iucGoToScreen));
        event->args.pushNew(makeIntValue(screen));
        ui.postEvent(event);
    }

    return ok;
}

/** Post a Go To Starchart request
    \param ui User interface binding
    \param pt Target position
    \return true iff request could be processed */
bool
postGoToChart(IntUserInterfaceBinding& ui, GPoint pt)
{
    // Prepare
    setStarchartPosition(pt);

    // Post the event
    std::auto_ptr<IntUserInterfaceEvent> event(new IntUserInterfaceEvent(iucGoToChart));
    ui.postEvent(event);

    return true;
}

/** Post a generic "Go to" request.
    \param ui User interface binding
    \param sel Selection object to use
    \param id Index for selection
    \param iuc User interface command
    \return true iff request could be processed */
static bool
postGoToGeneric(IntUserInterfaceBinding& ui, GObjectSelection& sel, int id, IntUserInterfaceCommand iuc)
{
    GObjectType* type = sel.getObjectType();
    if (type != 0 && trySetIndex(id, *type, sel)) {
        ui.postEvent(std::auto_ptr<IntUserInterfaceEvent>(new IntUserInterfaceEvent(iuc)));
        return true;
    } else {
        return false;
    }
}


/** Go to an object. Opens the control screen or the starmap, depending on the object.
    \param ui User interface binding
    \param obj Object to go to
    \param preferBases Prefer going to the starbase screen if possible
    \return true iff request could be processed */
bool
postGoToObject(IntUserInterfaceBinding& ui, GObject& obj, bool preferBases)
{
    if (GShip* sh = dynamic_cast<GShip*>(&obj)) {
        // Go to ship screen if played, show it in map otherwise
        return postGoToScreen(ui, 1, sh->getId())
            || postGoToChart(ui, sh->getPos());
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(&obj)) {
        // If this is a base query, and object is a base, go to base screen.
        // Otherwise, go to planet screen.
        return (preferBases && postGoToScreen(ui, 3, pl->getId()))
            || postGoToScreen(ui, 2, pl->getId())
            || postGoToChart(ui, pl->getPos());
    } else if (GMinefield* mf = dynamic_cast<GMinefield*>(&obj)) {
        // Minefield
        return postGoToGeneric(ui, current_minefield_selection, mf->getId(), iucGoToMinefield)
            || postGoToChart(ui, mf->getPos());
    } else if (GUfo* ufo = dynamic_cast<GUfo*>(&obj)) {
        // Ufo. Ufos cannot use the "Id = Index" equivalence.
        int index = findIndex(current_ufo_type, ufo);
        return index != 0
            && (postGoToGeneric(ui, current_ufo_selection, index, iucGoToUfo)
                || postGoToChart(ui, ufo->getPos()));
    } else if (GIonStorm* ion = dynamic_cast<GIonStorm*>(&obj)) {
        // Ion storm
        return postGoToGeneric(ui, current_ion_selection, ion->getId(), iucGoToStorm)
            || postGoToChart(ui, ion->getPos());
    } else if (GMapObject* mo = dynamic_cast<GMapObject*>(&obj)) {
        // Map object (i.e. anything)
        return postGoToChart(ui, mo->getPos());
    } else {
        // Not known
        return false;
    }
}
