/**
  *  \file client/tiles/histmovement.cc
  *  \brief WHistoryShipMovementTile tile
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "client/tiles/histmovement.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "game/ship.h"
#include "game/tables.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "util/nls.h"

/** Constructor.
    \param selection Object selection. Should be a WHistoryObjectSelection (but doesn't have to) */
WHistoryShipMovementTile::WHistoryShipMovementTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(GfxPoint(315, 3*font_heights[FONT_NORMAL]), selection),
      hos(dynamic_cast<WHistoryObjectSelection*>(&selection))
{
    if (hos) {
        conn_turn = observe(hos->sig_turn_changed);
    }
}

void
WHistoryShipMovementTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());

    UITableWriter tw(ctx, getInnerSize(5, 0), FONT_NORMAL);
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Distance:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Heading:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Warp factor:"));

    GShip* sh = dynamic_cast<GShip*>(getCurrentObject());
    if (sh) {
        tw.newLColumn();
        const int turn = hos ? hos->getTurn() : getCurrentTurn().getTurnNumber();
        const TDbShipTrackEntry& now  = sh->getHistoryEntry(turn);
        const TDbShipTrackEntry& prev = sh->getHistoryEntry(turn-1);

        /* Distance */
        bool notMoved = false;
        if (now.x > 0 && now.y > 0 && prev.x > 0 && prev.y > 0) {
            if (now.x == prev.x && now.y == prev.y) {
                notMoved = true;
                tw.writeVariableCell(UIColor::tc_Green, _("not moved"));
            } else {
                tw.writeVariableCell(UIColor::tc_Green,
                                     format(_("%.2f ly"),
                                            std::sqrt(double(distanceSquared(GPoint(now.x, now.y), GPoint(prev.x, prev.y))))));
            }
        } else {
            tw.writeVariableCell(UIColor::tc_Yellow, _("unknown"));
        }

        /* Heading. We rely on the loading code to have set this correctly. */
        if (now.heading >= 0) {
            tw.writeVariableCell(UIColor::tc_Green, format(_("%d\xC2\xB0 (%s)"),
                                                           now.heading,
                                                           getHeadingName(now.heading)));
        } else if (notMoved) {
            tw.writeVariableCell(UIColor::tc_Green, _("not moved"));
        } else {
            tw.writeVariableCell(UIColor::tc_Yellow, _("unknown"));
        }

        /* Speed */
        if (now.speed == 0 && notMoved) {
            tw.writeVariableCell(UIColor::tc_Green, _("not moved"));
        } else if (now.speed >= 0) {
            tw.writeVariableCell(UIColor::tc_Green, format(_("Warp %d"), now.speed));
        } else {
            tw.writeVariableCell(UIColor::tc_Yellow, _("unknown"));
        }
    }
    tw.finish();
}
