/**
  *  \file game/cargotr.cc
  *  \brief Cargo Transfer
  *
  *  This module implements cargo transfer.
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/cargotr.h"
#include "cpluslib/assert.h"
#include "util/nls.h"
#include "game/cargo.h"

/***************************** GCargoTransfer ****************************/


/** Constructor. */
GCargoTransfer::GCargoTransfer()
{ }

/** Destructor. Cancels the transaction if it was not yet finished. */
GCargoTransfer::~GCargoTransfer()
{
    if (!isFinished())
        doCancel();
}

/** Add a new container.
    The GCargoTransfer object will own the object, and control its lifetime.
    It will cancel or commit it. */
GCargoContainer&
GCargoTransfer::addNewContainer(GCargoContainer* obj)
{
    objs.push_back(obj);
    obj->sig_changed.add(&sig_changed, &Signal0::raise);
    return *obj;
}

/** Move stuff for 1:1 cargo transfer.
    \param type    type to move
    \param amount  amount to move. May be negative; semantics are
                   preserved.
    \param from    source object (0=left, 1=right)
    \param partial true to allow partial transfers, false to reject
    \param s2m     sell supplies on-the-fly
    \return amount moved.
    \inv !partial => return == 0 || return == amount
    \inv partial  => return >= 0 && return <= amount */
int32_t
GCargoTransfer::move(GCargoType type, int32_t amount, std::size_t from, bool partial, bool s2m)
{
    if (amount < 0) {
        return -moveFT(type, -amount, 1-from, from, partial, s2m);
    } else {
        return moveFT(type, amount, from, 1-from, partial, s2m);
    }
}

/** Move from/to.
    \param type Type to move
    \param amount Amount to move. Must be positive.
    \param from Source object
    \param to Destination object
    \param partial true to allow partial transfers, false to reject
    \param s2m     sell supplies on-the-fly
    \return amount moved. */
int32_t
GCargoTransfer::moveFT(GCargoType type, int32_t amount, std::size_t from, std::size_t to, bool partial, bool s2m)
{
    /* trivial case? */
    if (!amount || from == to)
        return 0;

    /* figure out src/dest objects and type */
    GCargoContainer& src  = *objs[from];
    GCargoContainer& dest = *objs[to];

    GCargoType dtype = (type == el_Supplies && s2m ? el_Money : type);

    /* can we actually move it? */
    if (!src.canHaveCargo(type) || !dest.canHaveCargo(dtype))
        return 0;

    int32_t space = dest.getMaxCargo(dtype) - dest.getCargo(dtype);
    if (space < 0)
        return 0;
    if (amount > space) {
        if (!partial)
            return 0;
        amount = space;
    }

    /* do the transfer */
    int32_t got = -src.changeCargo(type, -amount, partial);
    int32_t put = dest.changeCargo(dtype, got, partial);
    ASSERT(got == put);
    ASSERT(got <= amount);
    return got;
}

/** Perform unload. Cargo containers have a notion of "unload source"
    and "unload target" (can_Unload, can_GetUnload). If this
    GCargoTransfer object has containers of both types, this function
    performs the unload operation by attempting to move all minerals,
    colonists, supplies and cash from the source to the target. Fuel
    and ammunition are not moved as we assume that the source unit (a
    ship) needs them for itself.

    \param s2m true to sell supplies on-the-fly

    \return true iff we had both a source and a target object (does
    not say whether any cargo was moved, just whether the structural
    requirement was fulfilled) */
bool
GCargoTransfer::doUnload(bool s2m)
{
    /* figure out direction */
    if (objs.size() != 2) {
        return false;
    }

    std::size_t direction;
    if (objs[0]->hasFlag(GCargoContainer::can_Unload) && objs[1]->hasFlag(GCargoContainer::can_GetUnload)) {
        direction = 0;
    } else if (objs[1]->hasFlag(GCargoContainer::can_Unload) && objs[0]->hasFlag(GCargoContainer::can_GetUnload)) {
        direction = 1;
    } else {
        return false;
    }

    /* transfer */
    move(el_Tritanium,  objs[direction]->getCargo(el_Tritanium),  direction, true, s2m);
    move(el_Duranium,   objs[direction]->getCargo(el_Duranium),   direction, true, s2m);
    move(el_Molybdenum, objs[direction]->getCargo(el_Molybdenum), direction, true, s2m);
    move(el_Colonists,  objs[direction]->getCargo(el_Colonists),  direction, true, s2m);
    move(el_Supplies,   objs[direction]->getCargo(el_Supplies),   direction, true, s2m);
    move(el_Money,      objs[direction]->getCargo(el_Money),      direction, true, s2m);
    return true;
}

/** Check validity of cargo transfer.
    A transfer is only valid when all temporary spaces are empty. */
bool
GCargoTransfer::isValid() const
{
    for (std::size_t i = 0; i < objs.size(); ++i) {
        if (objs[i]->hasFlag(GCargoContainer::can_Temporary)) {
            if (objs[i]->isChanged()) {
                return false;
            }
        }
    }
    return true;
}

GError
GCargoTransfer::getError()
{
    if (!isValid()) {
        return GError(GError::ePerm, _("Hold spaces must be empty"));
    } else {
        return GError();
    }
}

void
GCargoTransfer::cancel()
{
    for (std::size_t i = 0; i < objs.size(); ++i)
        objs[i]->doCancel();
}

void
GCargoTransfer::commit()
{
    for (std::size_t i = 0; i < objs.size(); ++i)
        objs[i]->doCommit();
}
