/**
  *  \file game/gen.cc
  *  \brief "GEN" file management
  *
  *  This provides a class to parse GENx.DAT files and the respective
  *  RST section. The GEN file contains score data, passwords, and
  *  other administrative information.
  *
  *  (c) 2001-2015 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <algorithm>
#include "game/gen.h"
#include "cpluslib/assert.h"
#include "game/checksum.h"

/** "NOPASSWORD". Setting the password to this value (cleartext) will
    inhibit the password prompt in all programs. Note that "" (empty
    string) is a valid password, it will cause a prompt (which users
    confirm with Enter). */
const char GGen::NOPASSWORD[] = "NOPASSWORD";

/*! \class GGen
    \brief GEN file Parser

    This class provides functions to access TGen and TResultGen
    records, as read from GENx.DAT resp. RST files. Those records
    contain
    - checksums
    - scores
    - passwords
    - administrative information */

/** Initialize with GEN record (from GENx.DAT). */
GGen::GGen(const TGen& gen)
    : gen(gen)
{
    genSig();
}

/** Initialize with GEN record (from RST). */
GGen::GGen(const TResultGen& agen)
{
    std::memcpy(gen.timestamp, agen.timestamp, sizeof(gen.timestamp));
    std::memcpy(gen.scores,    agen.scores,    sizeof(gen.scores));
    gen.player_id = agen.player_id;
    std::memcpy(gen.password,  agen.password,  sizeof(agen.password));
    gen.unused = 0;
    gen.ship_checksum = agen.ship_checksum;
    gen.pdata_checksum = agen.pdata_checksum;
    gen.bdata_checksum = agen.bdata_checksum;
    gen.new_password_flag = 0;
    std::memset(gen.new_password, 0, sizeof(gen.new_password));
    gen.turn = agen.turn;
    gen.timestamp_checksum = agen.timestamp_checksum;
    genSig();
}

/** Fetch password, and store it in p. Returns length of password. p
    must point to 10 bytes. This routine is consistent with PCC 1.x in
    handling "bogus" passwords. This routine is deliberately not
    exported publicly. */
unsigned int
GGen::getPassword(char (&p)[10]) const
{
    if (gen.new_password_flag == 13) {
        /* New password has been set */
        for (int i = 0; i < 10; ++i)
            p[i] = gen.new_password[i] - 50;
    } else {
        /* No new password */
        for (int i = 0; i < 10; ++i) {
            uint8_t k = gen.password[i] - gen.password[19-i] + 32;
            if (k < 32 || k > 127)
                p[i] = 32;
            else
                p[i] = k;
        }
    }

    /* compute length of password */
    unsigned n = 10;
    while (n > 0 && p[n-1] <= 32)
        --n;
    return n;
}

/** Test password.
    \param pass password as provided by user
    \return true iff pass is the valid password. */
bool
GGen::isPassword(string_t pass) const
{
    char buffer[10];
    unsigned len = getPassword(buffer);
    return pass.length() == len && pass.compare(0, len, buffer, len) == 0;
}

/** Test whether password active.
    \return true iff this turn has an active password. */
bool
GGen::hasPassword() const
{
    char buffer[10];
    return !(getPassword(buffer) == 10 && std::memcmp(buffer, NOPASSWORD, 10) == 0);
}

/** Change password.
    \param pass password in clear text
    \pre pass.length() <= 10 */
void
GGen::setPassword(string_t pass)
{
    for (unsigned i = 0; i < 10; ++i)
        if (i >= pass.length())
            gen.new_password[i] = 50;
        else
            gen.new_password[i] = pass[i] + 50;
    gen.new_password_flag = 13;
}

/** Set password from TRN file.
    \param data password data as read from turn file (10 bytes, not null-terminated) */
void
GGen::setPasswordFromTurn(const char* data)
{
    std::copy(data, data + 10, gen.new_password);
    gen.new_password_flag = 13;
}

/** Get password to store in TRN file.
    \param p [out] Password data
    \retval true new password has been returned
    \retval false password was not changed, %p was not modified */
bool
GGen::getNewPasswordData(char (&p)[10]) const
{
    if (gen.new_password_flag == 13) {
        for (int i = 0; i < 10; ++i) {
            p[i] = gen.new_password[i];
        }
        return true;
    } else {
        return false;
    }
}

/** Compute signatures. */
void
GGen::genSig()
{
    for (int i = 0; i < 10; ++i)
        sig2[i] = gen.password[i+10] + i+1;
}

/** Get checksum. The player-side GEN file stores checksums; this
    routine returns the checksum for a particular area.

    The checksum is appropriate to the file this GGen object was constructed
    from; when constructed from a TGen, it's appropriate for a game directory
    (sum of all bytes in DAT+DIS including count and signature), when
    constructed from a TResultGen, it's appropriate for a result file
    (sum of all data bytes, not including count).

    \see setChecksum */
uint32
GGen::getChecksum(GControlSection sec) const
{
    switch (sec) {
     case ShipSection:   return gen.ship_checksum;
     case PlanetSection: return gen.pdata_checksum;
     case BaseSection:   return gen.bdata_checksum;
    }
    ASSERT(0);
}

/** Set checksum. Modifies the checksum fields that will be stored
    in the player-side GEN file.
    \param sec Section
    \param value Checksum, sum of all bytes (including count and signature)
    in the respective file. */
void
GGen::setChecksum(GControlSection sec, uint32 value)
{
    switch (sec) {
     case ShipSection:   gen.ship_checksum  = value; break;
     case PlanetSection: gen.pdata_checksum = value; break;
     case BaseSection:   gen.bdata_checksum = value; break;
    }
}

/** Validate timestamp checksum. This is the only checksum that validates the
    GEN record "as is" (in contrast to the other checksums, cf. getChecksum(),
    that validate whether the GEN record matches the game data.
    \return true iff valid */
bool
GGen::isValidTimestampChecksum() const
{
    return gen.timestamp_checksum == ::getChecksum(gen.timestamp, sizeof(gen.timestamp));
}

/** Make timestamp checksum valid.
    \post isValidTimestampChecksum() */
void
GGen::setTimestampChecksum()
{
    gen.timestamp_checksum = ::getChecksum(gen.timestamp, sizeof(gen.timestamp));
}
